/* 
   COPYRIGHT 2011 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - socklib.c */
/* rdk@krupczak.org */

/* a basic set of socket library routines attempting to isolate
   the messiness of BSD sockets; note that we do not do anything
   as of yet with IPv6; everything here assumes IPv4 XXX; 
   on Linux, everything needs to be in network-byte order a prior 
   see "man 7 ip" ; most functions appear to return things
   in network byte order as well

   support, on non-win32, for ipv6 address conversions added
   but not connecting/talking over ipv6 yet (3/21/2006)

   support for IPv6 added 9/21/2011 
*/


#ifdef WIN32
//#include <windows.h>
//#include <IPExport.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <stdio.h>
#endif

#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>

#ifndef WIN32
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <strings.h>
#endif

#ifdef SOLARIS
#include <sys/filio.h> /* for FIONREAD */
#endif

#ifndef WIN32
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#endif

#ifdef SOLARIS
#include <arpa/inet.h>
#include <inttypes.h>
#endif

#include "socklib.h"
#include "korgdebug.h"

#ifdef WIN32
#define bcopy(src,dst,n)  memcpy(dst,src,n)
#define bzero(dst,n)      memset(dst,0,n)
#endif

/* ****************************************************************** */
/* get a UDP/Internet socket and return it */

/* fortunately, in winsock, a SOCKET type is really an u_int */

int get_udp_socket()
{
     return socket(PF_INET,SOCK_DGRAM,IPPROTO_UDP);
}

int get_udp6_socket() { return socket(AF_INET6,SOCK_DGRAM,IPPROTO_UDP); }

/* ****************************************************************** */
/* get a TCP/Internet socket and return it */

int get_tcp_socket()
{
     return socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
}

int get_tcp6_socket() { return socket(AF_INET6,SOCK_STREAM,IPPROTO_TCP); }

/* ****************************************************************** */
/* bind a UDP socket to the specified port (network byte order) */
/* if reuseaddr=1, setsockopt to SO_REUSEADDR */

int bind_udp_socket(int s, u_int ipaddr, u_short port, int reuseaddr)
{
     struct sockaddr_in addr;
     int flag;

     if (reuseaddr) {
        flag = 1; 
        setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(void *)&flag,sizeof(flag));
     }

     addr.sin_family = PF_INET;
     addr.sin_port = port;
     addr.sin_addr.s_addr = ipaddr;

     return bind(s,(struct sockaddr *)&addr,sizeof(addr));
}

int bind_udp6_socket(int s, struct in6_addr ipaddr, u_short port, int reuseaddr)
{
     struct sockaddr_in6 addr;
     int flag;

     if (reuseaddr) {
        flag = 1; 
        if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(void *)&flag,sizeof(flag)) < 0)
	   perror("bind_udp6: setsockopt failed");
     }
     addr.sin6_family = PF_INET6;
     addr.sin6_port = port;
     addr.sin6_addr = ipaddr;

     return bind(s,(struct sockaddr *)&addr,sizeof(addr));
}

/* ****************************************************************** */
/* bind a TCP/Internet socket to the specified port (network byte order) */
/* if reuseaddr=1, setsockopt to SO_REUSEADDR */

int bind_tcp_socket(int s, u_int ipaddr, u_short port, int reuseaddr)
{
     struct sockaddr_in addr;
     int flag;

     if (reuseaddr) {
        flag = 1; 
        if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(void *)&flag,
                       sizeof(flag)) <0) {
 	   fprintf(stderr,"bind_tcp_socket: setsockop SO_REUSEADDR failed, %d",
                   errno);
        }
        else {
	   ; 
        }
     }

     addr.sin_family = PF_INET;
     addr.sin_port = port;
     addr.sin_addr.s_addr = ipaddr;

     return bind(s,(struct sockaddr *)&addr,sizeof(addr));
}

int bind_tcp6_socket(int s, struct in6_addr ipaddr, u_short port, int reuseaddr)
{
     struct sockaddr_in6 addr;
     int flag;

     if (reuseaddr) {
        flag = 1; 
        if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(void *)&flag,
                       sizeof(flag)) <0) {
 	   fprintf(stderr,"bind_tcp6_socket: setsockop SO_REUSEADDR failed, %d",
                   errno);
        }
        else {
	   ; 
        }
     }

     addr.sin6_family = PF_INET6;
     addr.sin6_port = port;
     addr.sin6_addr = ipaddr;

     return bind(s,(struct sockaddr *)&addr,sizeof(addr));
}

/* ****************************************************************** */
/* connect a TCP socket to dest addr, dest port (net byte order) */

int connect_tcp_socket(int s, u_int ipaddr, u_short port)
{
     struct sockaddr_in addr;

     addr.sin_family = PF_INET;
     addr.sin_port = port;
     addr.sin_addr.s_addr = ipaddr;

     return connect(s,(struct sockaddr *)&addr,sizeof(addr));
}

int connect_tcp6_socket(int s, struct in6_addr ipaddr, u_short port)
{
     struct sockaddr_in6 addr;
     socklen_t addrlen;

     bzero((void *)&addr,sizeof(addr));

     addr.sin6_family = PF_INET6;
     addr.sin6_port = port;
     addr.sin6_addr = ipaddr;
     addrlen = sizeof(addr);

     return connect(s,(struct sockaddr *)&addr,addrlen);
}

/* ****************************************************************** */
/* connect a TCP socket to dest addr/port (net byte order) waiting
   at most timeout seconds; if timeout or no connection, return -1; */

int connect_tcp6_socket_timeout(int s, struct in6_addr ipaddr, u_short port, 
                                int timeout)
{
     struct sockaddr_in6 addr;
     struct sockaddr_storage peeraddr;
     socklen_t addrlen;
     fd_set write_set;
     fd_set error_set;
     struct timeval timeoutval;
     int ret,on;
#ifdef WIN32
     int peernamelen;
#else
     socklen_t peernamelen;
#endif

     /* set socket non-blocking */
     on = 1;
#ifndef WIN32
     ioctl(s,FIONBIO,&on);
#else
     ioctlsocket(s,FIONBIO,&on);
#endif

     bzero((void *)&addr,sizeof(addr));

     addr.sin6_family = PF_INET6;
     addr.sin6_port = port;
     addr.sin6_addr = ipaddr;
     addrlen = sizeof(addr);

     ret = connect(s,(struct sockaddr *)&addr,sizeof(addr));

     /* DBG2("connect_tcp_socket_timeout: initial connect ret %d, errno:%d\n",ret,errno);*/

     if (ret == 0) {
       /*DBG("connect_tcp6_socket_timeout: immediate connection\n");*/
        /* immediate connection */
        return 1;
     }

     /* on solaris, this condition seems to occur; dont panic, 
        it appears "normal" proceed with async connection */
     if ((ret != 0) && (errno == 0)) {
        ; /* continue w/o error */
     }

#ifndef WIN32
     /* check for error and errno=EINPROGRESS */
     if ((errno != 0) && (errno != EINPROGRESS)) {
        /* DBG("connect_tcp_socket_maxdelay: error %d\n",errno);*/
        return -1;
     }
#else
     /* check for SOCK_ERROR and WSAGetLastError() = WSAEWOULDBLOCK */
#endif

     /* select for up to max seconds; if connection is available,
        then it will be available for writing */

     FD_ZERO(&write_set);
     FD_SET(s,&write_set);
     FD_ZERO(&error_set);
     FD_SET(s,&error_set);

     timeoutval.tv_sec = timeout;
     timeoutval.tv_usec = 0;
     
     if ((ret = select(s+1,NULL,&write_set,&error_set,&timeoutval)) <= 0) {
        /* timeout, connection not completed, nothing pending */
        /*DBG("connect_tcp6_socket_timeout: timeout or error\n");*/
        return -1;
     }

     if (FD_ISSET(s,&write_set) == 0) {
       /*DBG("connect_tcp6_socket_timeout: no connection\n");*/
        return -1;
     }

     if (FD_ISSET(s,&error_set) != 0) {
       /*DBG1("connect_tcp6_socket_timeout: error on socket %d\n",s);*/
        return -1;
     }

     /* in some cases (Solaris, Linux), we get this far but we really
        havent connected and we then generate SIGPIPE later on; test
        if connection succeeded by calling getpeername() */
     peernamelen = sizeof(peeraddr);
     if (getpeername(s,(struct sockaddr *)&peeraddr,&peernamelen) < 0) {
       /*DBG1("connect_tcp6_socket_timeout: getpeername failed, %d\n",errno);*/
        return -1;
     }

     /* DBG("connect_tcp6_socket_timeout: seems to have connected\n");*/

     /* set back to blocking? Yes, since that is the semantics
        that the rest of our socklib assumes */
     on = 0;
#ifndef WIN32
     ioctl(s,FIONBIO,&on);
#else
     ioctlsocket(s,FIONBIO,&on);
#endif

     return 1;
}

int connect_tcp_socket_timeout(int s, u_int ipaddr, u_short port, int timeout)
{
     struct sockaddr_in addr;
     struct sockaddr peeraddr;
     fd_set write_set;
     fd_set error_set;
     struct timeval timeoutval;
     int ret,on;
#ifdef WIN32
     int peernamelen;
#else
     socklen_t peernamelen;
#endif

     /* set socket non-blocking */
     on = 1;
#ifndef WIN32
     ioctl(s,FIONBIO,&on);
#else
     ioctlsocket(s,FIONBIO,&on);
#endif

     addr.sin_family = PF_INET;
     addr.sin_port = port;
     addr.sin_addr.s_addr = ipaddr;

     ret = connect(s,(struct sockaddr *)&addr,sizeof(addr));

     /* DBG2("connect_tcp_socket_timeout: initial connect ret %d, errno:%d\n",ret,errno);*/

     if (ret == 0) {
       /* DBG("connect_tcp_socket_timeout: immediate connection\n");*/
        /* immediate connection */
        return 1;
     }

     /* on solaris, this condition seems to occur; dont panic, 
        it appears "normal" proceed with async connection */
     if ((ret != 0) && (errno == 0)) {
        ; /* continue w/o error */
     }

#ifndef WIN32
     /* check for error and errno=EINPROGRESS */
     if ((errno != 0) && (errno != EINPROGRESS)) {
       /* DBG("connect_tcp_socket_maxdelay: error %d\n",errno);*/
        return -1;
     }
#else
     /* check for SOCK_ERROR and WSAGetLastError() = WSAEWOULDBLOCK */
#endif

     /* select for up to max seconds; if connection is available,
        then it will be available for writing */

     FD_ZERO(&write_set);
     FD_SET(s,&write_set);
     FD_ZERO(&error_set);
     FD_SET(s,&error_set);

     timeoutval.tv_sec = timeout;
     timeoutval.tv_usec = 0;
     
     if ((ret = select(s+1,NULL,&write_set,&error_set,&timeoutval)) <= 0) {
        /* timeout, connection not completed, nothing pending */
       /* DBG("connect_tcp_socket_timeout: timeout or error\n");*/
        return -1;
     }

     if (FD_ISSET(s,&write_set) == 0) {
       /* DBG("connect_tcp_socket_timeout: no connection\n"); */
        return -1;
     }

     if (FD_ISSET(s,&error_set) != 0) {
       /* DBG1("connect_tcp_socket_timeout: error on socket %d\n",s);*/
        return -1;
     }

     /* in some cases (Solaris, Linux), we get this far but we really
        havent connected and we then generate SIGPIPE later on; test
        if connection succeeded by calling getpeername() */
     peernamelen = sizeof(peeraddr);
     if (getpeername(s,(struct sockaddr *)&peeraddr,&peernamelen) < 0) {
       /* DBG1("connect_tcp_socket_timeout: getpeername failed, %d\n",errno);*/
        return -1;
     }

     /* DBG("connect_tcp_socket_timeout: seems to have connected\n"); */

     /* set back to blocking? Yes, since that is the semantics
        that the rest of our socklib assumes */
     on = 0;
#ifndef WIN32
     ioctl(s,FIONBIO,&on);
#else
     ioctlsocket(s,FIONBIO,&on);
#endif

     return 1;
}

/* ****************************************************************** */
/* connect UDP/Internet socket to destaddr, destport */

int connect_udp_socket(int s, u_int ipaddr, u_short port)
{
     struct sockaddr_in addr;

     addr.sin_family = PF_INET;
     addr.sin_port = port;
     addr.sin_addr.s_addr = ipaddr;

     return connect(s,(struct sockaddr *)&addr,sizeof(addr));
}

int connect_udp6_socket(int s, struct in6_addr ipaddr, u_short port)
{
     struct sockaddr_in6 addr;

     addr.sin6_port = port;
     addr.sin6_addr = ipaddr;

     return connect(s,(struct sockaddr *)&addr,sizeof(addr));
}

/* ****************************************************************** */
/* listen for a TCP connection request from some client */

int listen_for_tcp_connection(int s, int backlog)
{
     return listen(s,backlog);
}

int listen_for_tcp6_connection(int s, int backlog) { return listen(s,backlog);}

/* ****************************************************************** */
/* check to see if there is a connection attempt pending
   on socket s that is in listen state; wait for up to t seconds
   before returning; return 0 or 1 */

int connection_pending(int s, int timeout)
{
     fd_set fdset;
     struct timeval timeoutval;
     int ret;

     FD_ZERO(&fdset);
     FD_SET(s,&fdset);
     timeoutval.tv_sec = timeout;
     timeoutval.tv_usec = 0;

     ret = select(s+1,&fdset,NULL,NULL,&timeoutval);
     if (ret == 0) {
        /* timeout, nothing interesting, no connections pending */
        return 0;
     }
 
     /* we could check via FD_ISSET */

     return 1;
}

/* ****************************************************************** */
/* accept a TCP connection request and return new socket as return arg */
/* this function blocks until connection request */

int accept_tcp_connection_request(int s, u_int *ipaddr, u_short *port)
{
     struct sockaddr_in addr;
     int ret;
#ifdef WIN32
     int namelen;
#else
     socklen_t namelen;
#endif

     if ((ipaddr == NULL) || (port == NULL)) {
        fprintf(stderr,"accept_tcp_connection: invalid argument\n");
        return -1;
     }

     namelen = sizeof(addr);

     ret = accept(s,(struct sockaddr *)&addr,&namelen);

     *ipaddr = addr.sin_addr.s_addr;
     *port = addr.sin_port;

     return ret;
}

int accept_tcp6_connection_request(int s,struct in6_addr *ipaddr, u_short *port)
{
     struct sockaddr_in6 addr;
     int ret;
#ifdef WIN32
     int namelen;
#else
     socklen_t namelen;
#endif

     if ((ipaddr == NULL) || (port == NULL)) {
        fprintf(stderr,"accept_tcp_connection: invalid argument\n");
        return -1;
     }

     namelen = sizeof(addr);

     ret = accept(s,(struct sockaddr *)&addr,&namelen);

     *ipaddr = addr.sin6_addr;
     *port = addr.sin6_port;

     return ret;
}

/* ****************************************************************** */
/* given name, convert to ipV4addr -- grab first address in list */
/* in network byte order */

/* fix for reentrant versions of gethostbyname!! */

#if defined(SOLARIS) || defined(LINUX)

u_int hostname_to_ipaddr(char *hostname)
{
     struct hostent h, *hptr;
     u_int addr;
     char buf[8192];
     int herrno;

     if (hostname == NULL) {
        return (u_int)0;
     }

     hptr = NULL;

     addr = inet_addr(hostname);

     if (addr == (u_int)-1) {

#ifdef LINUX
     if (gethostbyname_r(hostname,&h,buf,sizeof(buf),&hptr,&herrno) != 0) {
         /* DBG("hostname_to_ipaddr: gethostbyname_r != 0\n");*/
         return 0;
     }
#endif
#ifdef SOLARIS
       if ((hptr = gethostbyname_r(hostname,&h,buf,sizeof(buf),
                                    &herrno)) == NULL) {
	   /* DBG("hostname_to_ipaddr: gethostbyname_r hptr = null\n");*/
           return 0;
        }
#endif
        if (hptr == NULL) {
	   /* DBG("hostname_to_ipaddr: hptr = null\n");*/
           return 0;
        }
        if (hptr->h_length != 4) {
           fprintf(stderr,"hostname_to_ipaddr: address of length %d\n",
                   hptr->h_length);
           return 0;
        }
        if (hptr->h_addr == NULL) {
	   /* DBG("hostname_to_ipaddr: h_addr is null\n"); */
	   return 0;
        }

        addr = *((u_int *)(hptr->h_addr));

     } /* inet_addr == -1 */

     /*DBG1("hostname_to_ipaddr: returning %x\n",addr);*/

     return addr;
}

#endif /* SOLARIS || LINUX */

#ifdef WIN32

u_int hostname_to_ipaddr(char *hostname)
{
     struct hostent *h;
     u_int addr;

     if (hostname == NULL) {
        return (u_int)0;
     }

     addr = inet_addr(hostname);
     if (addr == (u_int)-1) {

        if ((h = gethostbyname(hostname)) == NULL) {
#if 0
#ifdef LINUX
            herror("gethostbyname failed");
#else
            perror("gethostbyname failed");
#endif
           fprintf(stderr,". . . .for %s\n",hostname);
#endif /* 0 */
           return (u_int)0;
        }

        if (h->h_length != 4) {
           fprintf(stderr,"hostname_to_ipaddr: address of length %d\n",
                   h->h_length);
           return 0;
        }
        if (h->h_addr == NULL) {
	  /*fprintf(stderr,"hostname_to_ipaddr: h_addr is null\n"); */
	   return 0;
        }

        addr = *((u_int *)(h->h_addr));

     } /* inet_addr == -1 */

     return addr;
}

#endif /* WIN32 */

/* ****************************************************************** */
/* take a hostname and return IPv6 addr if possible; return 1 if
   possible, < 0 etc. if not; 
   do we return the first IPv6 addr we found regardless of scope or
   try to find a particular addr??  For now, return first match
 */

int hostname_to_ipv6addr(char *hostname, struct in6_addr *retaddr)
{
    struct addrinfo *addrList;
    struct addrinfo *addr;
    struct addrinfo hints;
    struct sockaddr_in6 *sa6;
    int ret;

    bzero((void *)&hints,sizeof(hints));
    hints.ai_family = AF_INET6;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    addrList = NULL;
    if ((ret = getaddrinfo(hostname,NULL,&hints,&addrList)) < 0) {
       /* sometimes, we can get ret < 0 but no error; this happens
          when there are no matching answers; if this is the case,
          punt and return 0 */
       /* perror("getaddrinfo failed");*/
       DBG1("getaddrinfo failed '%s'\n",gai_strerror(ret));
    }

    /* on solaris, we can apparently get no error and null addrList on
       return; check for this condition */
    if (addrList == NULL) {
       DBG("hostname_to_ipv6addr: null addrList\n");
       return -2;
    }

    for (addr = addrList; addr != NULL; addr = addr->ai_next) {

        if (addr->ai_family == AF_INET6) {

           /* ai_addr should be sockaddr_in6 */
           sa6 = (struct sockaddr_in6 *)addr->ai_addr;
           if (addr->ai_addrlen == sizeof(struct sockaddr_in6)) {
              *retaddr = sa6->sin6_addr;
              if (addrList != NULL) freeaddrinfo(addrList);
              return 1;
           }
        }

        if (addr->ai_family == AF_INET) {
        }
    }

    if (addrList != NULL)
       freeaddrinfo(addrList);
  
    /* did not find an IPv6 address */
    return -1;
}

/* ****************************************************************** */
/* given an IPv4 address (in network byte order), convert and
   print via the inet_ntoa */

int convert_ipaddr(u_int ipaddr, char *ipaddrstr, int maxlen)
{
     char *str;
     struct in_addr in;

     if ((ipaddrstr == NULL) || (maxlen < 8)) {
        return -1;
     }

     in.s_addr = ipaddr;
     str = inet_ntoa(in);

     strncpy(ipaddrstr,str,maxlen-1);

     return 1;
}

/* ****************************************************************** */
/* given an IPv6 address, convert to string */

int convert_ipv6addr(u_char *ipv6addr, char *ipv6str, int maxlen)
{
     struct in6_addr *addr6;

     if ((ipv6str == NULL) || (maxlen < 32)) {
        return -1;
     }

     ipv6str[0] = 0;

     /* passed in as array of or ptr to bytes */
     addr6 = (struct in6_addr *)ipv6addr;

     /* print 4-char chunks at a time as per RFC 2373 */
#if 1
     sprintf(ipv6str,"%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
             addr6->s6_addr[0],
             addr6->s6_addr[1],
	     addr6->s6_addr[2],
	     addr6->s6_addr[3],
	     addr6->s6_addr[4],
	     addr6->s6_addr[5],
	     addr6->s6_addr[6],
	     addr6->s6_addr[7],
	     addr6->s6_addr[8],
	     addr6->s6_addr[9],
	     addr6->s6_addr[10],
	     addr6->s6_addr[11],
	     addr6->s6_addr[12],
	     addr6->s6_addr[13],
	     addr6->s6_addr[14],
	     addr6->s6_addr[15]);
#endif
#if 0
     sprintf(ipv6str,"%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x",
             ntohs(addr6->s6_addr16[0]),
             ntohs(addr6->s6_addr16[1]),
             ntohs(addr6->s6_addr16[2]),
             ntohs(addr6->s6_addr16[3]),
             ntohs(addr6->s6_addr16[4]),
             ntohs(addr6->s6_addr16[5]),
             ntohs(addr6->s6_addr16[6]),
             ntohs(addr6->s6_addr16[7]));
#endif
#if 0
     sprintf(ipv6str,"%08x:%08x:%08x:%08x",
             ntohl(addr6->s6_addr32[0]),
             ntohl(addr6->s6_addr32[1]),
             ntohl(addr6->s6_addr32[2]),
             ntohl(addr6->s6_addr32[3]));
#endif

#if 0
     inet_ntop(AF_INET6,addr6,ipv6str,maxlen);
#endif

     return 1;
}

/* ****************************************************************** */
/* 
   keep reading, from socket s, into buf[maxlen] until X bytes are 
   received or until some recv error
*/
  
int read_upto_X_bytes(int s, u_char *buf, int X, int maxlen)
{
     int retlen = 0;
     int ret;

     if (buf == NULL) {
        return -1;
     }
     if (maxlen < X) {
        fprintf(stderr,"read_upto_X_bytes: not enough space, %d < %d\n",
                maxlen,X);
        return -1;
     }

     while ((X - retlen) > 0) {

       if ((ret = recv(s,(void *)&buf[retlen],X-retlen,0)) < 0)
 	  break;

       retlen += ret;
     }

     /* printf("read_upto_X_bytes: read %d bytes\n",retlen); */

     return retlen;
}

/* ****************************************************************** */
/* check to see if data at socket to read, and how much */

#ifndef WIN32

/* UNIX */

int bytes_at_socket(int s)
{
     int bytes;

     if (s == 0) {
        return 0;
     }

     bytes = 0;
     if (ioctl(s,FIONREAD,(char *)&bytes) < 0)
        perror("ioctl FIONREAD failed");

     return bytes;
}

#else /* WIN32 */

int bytes_at_socket(int s)
{
     u_long bytes;

     if (s == 0) {
        return 0;
     }

     bytes = 0;

     if (ioctlsocket(s,FIONREAD,(u_long *)&bytes) != 0)
        perror("ioctl FIONREAD failed");

     return (int)bytes;
}

#endif /* !WIN32 */

/* ****************************************************************** */
/* given a TCP socket, turn off buffering so that all bytes get sent
   when we call send -- however, setting this still does not guarantee 
   that TCP can deliver on message boundaries as TCP fundamentally is 
   a byte-stream interface, not a message interface 
*/

int set_tcp_no_delay(int s)
{
     int ret;
     int nodelay = 1;

     if ((ret = setsockopt(s,IPPROTO_TCP,TCP_NODELAY,(char *)&nodelay,
                           sizeof(nodelay))) < 0)
        perror("setsockopt: tcp_nodelay failed");

     return ret;
 
}

/* ****************************************************************** */
/* set a socket's send/receive buffer sizes -- UDP or TCP */
/* if either size is 0, dont set that buffer's size */

int set_socket_buffer_sizes(int s, int recvsize, int sendsize)
{
     if (recvsize > 0) {
        if (setsockopt(s,SOL_SOCKET,SO_RCVBUF,(char *)&recvsize,sizeof(int))<0)
	   perror("setsockopt of SO_RCVBUF failed");
     }

     if (sendsize > 0) {
        if (setsockopt(s,SOL_SOCKET,SO_SNDBUF,(char *)&sendsize,sizeof(int))<0)
	   perror("setsockopt of SO_SNDBUF failed");
     }

     return 1;
}

/* ****************************************************************** */
/* get socket buffer send/receive size -- UDP or TCP socket */
/* if null ptr for either, dont retrieve/return that one */

int get_socket_buffer_sizes(int s, int *recvsize, int *sendsize) 
{
     int sndbufsz, rcvbufsz;
#ifdef WIN32
     int optlen;
#else
     socklen_t optlen;
#endif

     if (recvsize != NULL) {
        *recvsize = -1;
        optlen = sizeof(int);
        if (getsockopt(s,SOL_SOCKET,SO_RCVBUF,(char *)&rcvbufsz,&optlen) >= 0)
 	   *recvsize = rcvbufsz;
        else
	   perror("getsockopt of SO_RCVBUF failed");
     } 

     if (sendsize != NULL) {
        *sendsize = -1;
        optlen = sizeof(int);
        if (getsockopt(s,SOL_SOCKET,SO_SNDBUF,(char *)&sndbufsz,&optlen) >= 0)
 	   *sendsize = sndbufsz;
        else
	   perror("getsockopt of SO_SNDBUF failed");
     } 

     return 1;
}

/* ****************************************************************** */
/* given a address/string (hostname or IP address), convert it
   to a (hopefully FQDN) hostname of max length maxlen and return */

/* update these using reentrant fuctions !!! */
/* update for IPv6?; no, caller should use ipAddrToHostname (1 or 2) */

#if defined(SOLARIS) || defined(LINUX)

int addressToHostname(char *addrstr, char *hostname, int maxlen)
{
     struct in_addr a;
     struct hostent h, *hptr;
     char buf[8192];
     int herrno;

     if ((addrstr == NULL) || (hostname == NULL) || (maxlen < 5))
        return -1;

     /*DBG1("addressToHostname: resolving %s\n",addrstr);*/

     /* default, in case of error, is simply to return what 
        was passed in 
      */
     hostname[0] = 0;
     strncpy(hostname,addrstr,maxlen);
     hostname[maxlen-1] = 0;

     a.s_addr = 0; /* default in case of errors */
     bzero((void *)&h,sizeof(h));
     buf[0] = 0;

     /* get matching address; if its a hostname,
        0 returned and we try to get address 
      */ 
#ifdef SOLARIS
     /* inet_aton() not supported on win32 and Solaris */
#define INADDR_NONE -1

     if ((a.s_addr = inet_addr(addrstr)) == INADDR_NONE) {

       if ((hptr = gethostbyname_r(addrstr,&h,buf,sizeof(buf),&herrno)) == NULL) {
	 /*DBG("addressToHostname: gethostbyname_r hptr = null\n");*/
           return 1;
        }
        /* check to make sure h->h_addr is valid */
        if (hptr->h_addr) {
           a.s_addr = *((u_int *)(hptr->h_addr));
        }
        else {
	  /*DBG("addressToHostname: gethostbyname_r no addr\n");*/
	   return -1;
        }
     }

     hptr = NULL;
     if ((hptr = gethostbyaddr_r((char *)&a,sizeof(a),
				 AF_INET,&h,buf,
                                 sizeof(buf),&herrno)) == NULL) {

       /*DBG("addressToHostname: gethostbyaddr_r hptr = null\n");*/
        return 1;
     }

#else /* LINUX */
     if (inet_aton(addrstr,&a) == 0) {

        hptr = NULL;
        if (gethostbyname_r(addrstr,&h,buf,sizeof(buf),&hptr,&herrno) != 0) {
	  /*DBG("addressToHostname: gethostbyname_r != 0\n");*/
           return 1;
        }
        if (hptr == NULL) {
	  /*DBG("addressToHostname: gethostbyname_r hptr = null\n");*/
           return 1;
        }
        /* check to make sure h->h_addr is valid */
        if (hptr->h_addr) {
           a.s_addr = *((u_int *)(hptr->h_addr));
        }
        else {
	  /*DBG("addressToHostname: no h_addr\n");*/
	   return -1;
        }
     }

     hptr = NULL;
     herrno = 0;

     /* DBG2("xmpAddressToHostname: addrstr=%s a.addr = %lx\n",addrstr,a.s_addr); */

     if (gethostbyaddr_r((char *)&a,sizeof(a),
                          AF_INET,&h,buf,sizeof(buf),
			 &hptr,&herrno) != 0) {
       /* DBG1("addressToHostname: gethostbyaddr_r != 0, %d\n",herrno); */
        return 1;
     }
#endif

     /* now, hptr points to our returned struct */
     if (hptr == NULL) {
       /* DBG("addressToHostname: null hptr\n"); */
        return 1;
     }
   
     if (hptr->h_name == NULL) {
       /* DBG("addressToHostname: null h_name\n"); */
        return 1;
     }

     strncpy(hostname,hptr->h_name,maxlen);
     hostname[maxlen-1] = 0;
 
     /* DBG2("addressToHostname: finished %s -> %s\n",addrstr,hostname); */

     return 1;
}

#endif /* SOLARIS || LINUX */

#ifdef WIN32

int addressToHostname(char *addrstr, char *hostname, int maxlen)
{
     struct in_addr a;
     struct hostent *h;

     if ((addrstr == NULL) || (hostname == NULL) || (maxlen < 5))
        return -1;

     /* default, in case of error, is simply to return what 
        was passed in 
      */
     hostname[0] = 0;
     strncpy(hostname,addrstr,maxlen);
     hostname[maxlen-1] = 0;

     a.s_addr = 0; /* default in case of errors */

     /* get matching address; if its a hostname,
        0 returned and we try to get address 
      */ 

     if ((a.s_addr = inet_addr(addrstr)) == INADDR_NONE) {

        if ((h = gethostbyname(addrstr)) == NULL) {
           return 1;
        }
        /* check to make sure h->h_addr is valid */
        if (h->h_addr) {
           a.s_addr = *((u_int *)(h->h_addr));
        }
        else {
	   return -1;
        }
     }

     if ((h = gethostbyaddr((char *)&a,sizeof(a),AF_INET)) == NULL) {
        return 1;
     }
     if (h->h_name == NULL) {
        return 1;
     }

     strncpy(hostname,h->h_name,maxlen);
     hostname[maxlen-1] = 0;

     return 1;
}

#endif /* WIN32 */

/* ****************************************************************** */
/* all platforms: Solaris, Linux, win32 */
/* take: ipv4 addr, ipv6addr, partial hostname, fqdn, return fqdn */

int ipAddrToHostname(char *addrstr, char *hostname, int maxlen)
{
     struct addrinfo *result;
     struct addrinfo *res;
     int ret;

     if ((addrstr == NULL) || (hostname == NULL) || (maxlen < 5))
        return -1;

     /* if we already have a hostname, should we not just return it ? 
        Well, we want a FQDN and we dont know if we have a FQDN; the
        reason we are asking these questions is we are having problems
        with hostnames like time.windows.com which goes to an addr
        but resolves back to something else each time we call; thus
        we get different answers back and this leads to all kinds of
        problems */

     /* in case of error, simply return what was passed in */
     result = NULL;
     hostname[0] = 0;
     strncpy(hostname,addrstr,maxlen);
     hostname[maxlen-1] = 0;

     ret = getaddrinfo(addrstr,NULL,NULL,&result);
    
     if (ret != 0) {
       /* DBG2("ipAddrToName: getaddrinfo of %s error %s\n",addrstr,gai_strerror(ret)); */
        return -1;
     }

     for (res = result; res != NULL; res = res->ai_next) {

       ret = getnameinfo(res->ai_addr,res->ai_addrlen,hostname,maxlen,NULL,0,0);
       if (ret == 0) {
	  /* fprintf(stderr,"ipAddrToName: %s -> %s\n",addrstr,hostname); */
          freeaddrinfo(result);
          return 1;
       }
       else {
	 /* DBG2("ipAddrToName: getnameinfo %s error %s\n",addrstr,gai_strerror(ret)); */
          /* continue to see if we get an answer */
       }
     }

     freeaddrinfo(result);

     return 1;
}

/* ****************************************************************** */
/* all platforms: Solaris, Linux, win32 */
/* take: ipv4 addr, ipv6addr, partial hostname, fqdn, return fqdn */

/* valgrind insists there is a memory leak with our first call
   to getaddrinfo() despite our valient attempts to free everything
   we can find XXX */

int ipAddrToHostname2(char *addrstr, char *hostname, int maxlen)
{
     struct addrinfo *result1,*result2,*result3;
     struct addrinfo *res;
     struct addrinfo hints;
     int ret;

     if ((addrstr == NULL) || (hostname == NULL) || (maxlen < 5))
        return -1;

     /* if we already have a hostname, should we not just return it ? 
        Well, we want a FQDN and we dont know if we have a FQDN; the
        reason we are asking these questions is we are having problems
        with hostnames like time.windows.com which goes to an addr
        but resolves back to something else each time we call; thus
        we get different answers back and this leads to all kinds of
        problems; 
  
        As a new and improved algorithm, first test if we have
        an IP address; if its not an ip address, re-query and
        take the canonical name; if it is an ip address,
        re-query and then reverse lookup; this helps us
        consistently resolve time.windows.com to time.microsoft.akadns.net
     */

     /* in case of error, simply return what was passed in */
     hostname[0] = 0;
     strncpy(hostname,addrstr,maxlen);
     hostname[maxlen-1] = 0;

     /* test if this is an IP address; if it is not, we'll get an error */

     result1 = result2 = result3 = NULL;

     bzero((void *)&hints,sizeof(hints));
     hints.ai_flags = AI_CANONNAME | AI_NUMERICHOST;
     ret = getaddrinfo(addrstr,NULL,&hints,&result1);

     if (ret != 0) {
        bzero((void *)&hints,sizeof(hints));
        hints.ai_flags = AI_CANONNAME;
        ret = getaddrinfo(addrstr,NULL,&hints,&result2);
        if (ret != 0) {
 	   /* fprintf(stderr,"ipAddrToHostname2: getaddrinfo2 error %s\n",
	      gai_strerror(ret)); */
           if (result1 != NULL) freeaddrinfo(result1);
           if (result2 != NULL) freeaddrinfo(result2);
           return -1;
        }
        if (result2->ai_canonname != NULL) {
        /*  DBG1("ipAddrToHostname2: canonname is %s\n",result->ai_canonname);*/
           strncpy(hostname,result2->ai_canonname,maxlen);
           hostname[maxlen-1] = 0;
           if (result1 != NULL) freeaddrinfo(result1);
           if (result2 != NULL) freeaddrinfo(result2);
           return 1;
        }

     }

     /* if we passed in an address string, canonical name is 
        the address ip address; 
        if we passed in a partial name, we get back the 
        FQDN */
  
     result3 = result1;
     if (result2 != NULL)
        result3 = result2;

     for (res = result3; res != NULL; res = res->ai_next) {

         ret = getnameinfo(res->ai_addr,res->ai_addrlen,hostname,maxlen,NULL,0,0);
         if (ret == 0) {
            if (result1 != NULL) freeaddrinfo(result1);
            if (result2 != NULL) freeaddrinfo(result2);
            return 1;
         }
         else {
   	   /* DBG2("ipAddrToName2: getnameinfo error resolving %s %s\n",
	      addrstr,gai_strerror(ret)); */
           /* continue to see if we get an answer */
         }

     }

     if (result1 != NULL) freeaddrinfo(result1);
     if (result2 != NULL) freeaddrinfo(result2);

     return 1;
}

/* ****************************************************************** */
/* given an ip address string, return the IP version 4, or 6; 0 if error */

int ipVersionOfAddress(char *addrstr)
{
     struct sockaddr_storage sas;
     int addrlen;

     if (addrstr == NULL) return 0;

#ifndef WIN32
     struct in_addr v4addr;
     struct in6_addr v6addr;

     /* check if its an IPv4 address */
     if (inet_pton(AF_INET,addrstr,(void *)&v4addr) == 1) return 4;

     /* try to see if its a v6 address */
     if (inet_pton(AF_INET6,addrstr,(void *)&v6addr) == 1) return 6;
#else
     addrlen = sizeof(sas);

     if (WSAStringToAddress(addrstr,AF_INET,NULL,(struct sockaddr *)&sas,&addrlen) == 0) return 4;

     addrlen = sizeof(sas);
     if (WSAStringToAddress(addrstr,AF_INET6,NULL,(struct sockaddr *)&sas,&addrlen) == 0) return 6;
#endif

     return 0;
}

/* ****************************************************************** */
/* given a socket, get its ip address and port; return in str; return 4,
   6, or 0 for error */

int getMyAddress(int s, char *addrstr, int maxlen)
{
     int ret;
     struct sockaddr_storage sas;
     socklen_t addrlen = sizeof(sas);
     char ipaddr[128];
     struct sockaddr_in6 *sa6;
     struct sockaddr_in *sa;
    
     if ((addrstr == NULL) || (maxlen == 0)) 
        return 0;

     addrstr[0] = 0;

     if ((ret = getsockname(s,(struct sockaddr *)&sas,&addrlen)) < 0) {
        perror("getMyAddress: getsockname failed");
        return 0;
     }

     if (sas.ss_family == AF_INET6) {
        sa6 = (struct sockaddr_in6 *)&sas;
        convert_ipv6addr((u_char *)&sa6->sin6_addr,ipaddr,128);
        sprintf(addrstr,"[%s]:%d\n",
                ipaddr,ntohs(sa6->sin6_port));
        return 6;

     }
     if (sas.ss_family == AF_INET) {
        sa = (struct sockaddr_in *)&sas;
        convert_ipaddr(sa->sin_addr.s_addr,ipaddr,128);
        sprintf(addrstr,"%s:%d\n",
                ipaddr,ntohs(sa->sin_port));
        return 4;
     }

     return 0;
}


/* ****************************************************************** */
/* given a socket, return 4 or 6 indicating which socket family 
   its bound to; in recent version of Linux, we can use SO_DOMAIN 
   socket option */

int ipVersionOfSocket(int s)
{
     int ret;
     struct sockaddr_storage sas;
     socklen_t addrlen = sizeof(sas);
     
     if ((ret = getsockname(s,(struct sockaddr *)&sas,&addrlen)) < 0) {
        perror("ipVersionOfSocket: getsockname failed");
        return 0;
     }

     if (sas.ss_family == AF_INET6)
        return 6;
     if (sas.ss_family == AF_INET)
        return 4;

     fprintf(stderr,"ipVersionOfSocket: unknown ip version %d\n",sas.ss_family);
     return 0;
}

/* ****************************************************************** */
